/*--------------------------------------------------------------------------+
| Copyright 2021 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.variability.ui.properties;

import static org.eclipse.jface.layout.GridDataFactory.fillDefaults;
import static org.eclipse.jface.window.Window.OK;
import static org.eclipse.swt.SWT.FILL;
import static org.eclipse.swt.SWT.FLAT;
import static org.eclipse.swt.SWT.NONE;
import static org.fortiss.variability.model.features.CrossFeatureConstraintType.EXCLUDES;
import static org.fortiss.variability.model.features.CrossFeatureConstraintType.REQUIRES;
import static org.fortiss.variability.util.VariabilityUtils.getAllLegalConstraintTargets;

import java.util.List;

import org.conqat.ide.commons.ui.jface.TreeContentProviderBase;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ComboViewer;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import org.fortiss.tooling.base.ui.dialog.ElementTreeSingleSelectDialog;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.ui.extension.base.PropertySectionBase;
import org.fortiss.tooling.kernel.ui.presentation.ModelElementLabelProvider;
import org.fortiss.variability.model.features.AbstractCrossFeatureConstraint;
import org.fortiss.variability.model.features.AbstractFeature;
import org.fortiss.variability.model.features.CrossFeatureConstraintType;

/**
 * Property Section for {@link AbstractCrossFeatureConstraint}s.
 * 
 * @author bayha
 */
public class CrossFeatureConstraintPropertySectionBase extends PropertySectionBase {

	/** {@link ComboViewer} to select type. */
	ComboViewer typeCombo;

	/** Viewer for the target Feature. */
	TableViewer targetFeatureViewer;

	/** The {@link AbstractCrossFeatureConstraint} to display properties for. */
	AbstractCrossFeatureConstraint constraint;

	/** {@inheritDoc} */
	@Override
	protected void setSectionInput(Object input) {
		if(input instanceof AbstractCrossFeatureConstraint) {
			constraint = (AbstractCrossFeatureConstraint)input;
			if(typeCombo != null) {
				typeCombo.setInput(constraint);
				typeCombo.setSelection(new StructuredSelection(constraint.getType()));
			}
			if(targetFeatureViewer != null) {
				targetFeatureViewer.setInput(constraint);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void createControls(Composite parent, TabbedPropertySheetPage tabbedPropertySheetPage) {
		super.createControls(parent, tabbedPropertySheetPage);

		createTypeCombo();

		createTargetFeatureSelection();
	}

	/** Creates the Combo for selection of Constraint type. */
	protected void createTypeCombo() {
		typeCombo = createFormComboViewer("Constraint Type");
		typeCombo.setContentProvider(new ArrayContentProvider() {
			@Override
			public Object[] getElements(Object inputElement) {
				return new Object[] {REQUIRES, EXCLUDES};
			}
		});
		typeCombo.setLabelProvider(new ModelElementLabelProvider());
		typeCombo.addSelectionChangedListener(new ISelectionChangedListener() {

			@Override
			public void selectionChanged(SelectionChangedEvent event) {
				if(event.getSelection() instanceof StructuredSelection) {
					final StructuredSelection ss = (StructuredSelection)event.getSelection();
					if(ss.getFirstElement() instanceof CrossFeatureConstraintType) {
						final CrossFeatureConstraintType newType =
								(CrossFeatureConstraintType)ss.getFirstElement();

						// Change type only if there actually was a change
						if(constraint.getType() != newType) {
							ICommandStackService.getInstance().runAsCommand(constraint,
									new Runnable() {

										@Override
										public void run() {
											constraint.setType(newType);
										}
									});
						}
					}
				}
			}
		});
	}

	/** Creates the controls for the target feature selection. */
	protected void createTargetFeatureSelection() {
		final Composite featureComposite = new Composite(composite, NONE);
		featureComposite.setLayout(new GridLayout(2, false));
		fillDefaults().grab(true, false).applyTo(featureComposite);

		targetFeatureViewer = new TableViewer(featureComposite, FLAT);
		targetFeatureViewer.setContentProvider(new ArrayContentProvider() {
			@Override
			public Object[] getElements(Object inputElement) {
				if(inputElement instanceof AbstractCrossFeatureConstraint) {
					AbstractFeature f = ((AbstractCrossFeatureConstraint)inputElement).getTarget();
					if(f != null) {
						return new Object[] {f};
					}
				}
				return new Object[0];
			}
		});
		targetFeatureViewer.setLabelProvider(new ModelElementLabelProvider());
		targetFeatureViewer.setInput(constraint);

		Button selectButton = getWidgetFactory().createButton(featureComposite, "...", NONE);
		selectButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				List<AbstractFeature> features = getAllLegalConstraintTargets(constraint);
				ElementTreeSingleSelectDialog dialog = new ElementTreeSingleSelectDialog(
						featureComposite.getShell(), features, null, new TreeContentProviderBase() {

							@Override
							public Object[] getChildren(Object parentElement) {
								return new Object[0];
							}
						}, new ModelElementLabelProvider(), null);
				if(dialog.open() == OK) {
					final EObject selectedElement = (EObject)dialog.getSelectedElement();
					if(selectedElement instanceof AbstractFeature) {
						ICommandStackService.getInstance().runAsCommand(constraint, new Runnable() {
							@Override
							public void run() {
								constraint.setTarget((AbstractFeature)selectedElement);
								targetFeatureViewer.refresh();
							}
						});
					}
				}
			}
		});

		GridData layoutData = new GridData(FILL, FILL, true, false, 1, 1);
		layoutData.heightHint = selectButton.getSize().y;
		targetFeatureViewer.getTable().setLayoutData(layoutData);

		createFormEntry(featureComposite, "Target Feature");
	}
}
