/*-------------------------------------------------------------------------+
| Copyright 2022 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.variability.analysis;

import java.util.HashMap;
import java.util.Map;

/**
 * A map which uses a combination of two keys to identify all values.
 * 
 * @author bayha
 */
public class DualKeyMap<K1, K2, V> {
	/** Internally used map of maps. */
	private Map<K1, Map<K2, V>> internalMap = new HashMap<K1, Map<K2, V>>();

	/**
	 * Retrieves the value for the given combination of keys.
	 * 
	 * @param key1
	 *            The first part of the key.
	 * @param key2
	 *            The second part of the key.
	 * 
	 * @return The value which is identified by the given combination of the two keys.
	 */
	public V get(K1 key1, K2 key2) {
		Map<K2, V> map = internalMap.get(key1);
		if(map != null) {
			return map.get(key2);
		}
		return null;
	}

	/**
	 * Adds the given value for the given combination of keys.
	 * 
	 * @param key1
	 *            The first part of the key.
	 * @param key2
	 *            The second part of the key.
	 * @param value
	 *            The value to be associated with the given combination of keys.
	 * 
	 * @return The previous value associated with the given combination of keys, or null if there
	 *         was no mapping.
	 */
	public V put(K1 key1, K2 key2, V value) {
		Map<K2, V> map = internalMap.get(key1);
		if(map == null) {
			map = new HashMap<K2, V>();
			internalMap.put(key1, map);
		}

		return map.put(key2, value);
	}
}
