package org.fortiss.variability.analysis;

import static java.util.stream.Collectors.toList;
import static org.eclipse.emf.ecore.util.EcoreUtil.copy;
import static org.fortiss.variability.model.VariabilityModelElementFactory.createAndPC;
import static org.fortiss.variability.util.VariabilityUtilsInternal.getAllReferences;
import static org.fortiss.variability.util.VariabilityUtilsInternal.getChildrenWithType;
import static org.fortiss.variability.util.VariabilityUtilsInternal.logError;
import static org.fortiss.variability.util.VariabilityUtilsInternal.logInfo;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EEnumLiteral;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EStructuralFeature.Setting;
import org.eclipse.emf.ecore.EcorePackage;
import org.fortiss.variability.model.IAlternative;
import org.fortiss.variability.model.IOptionalVariationPoint;
import org.fortiss.variability.model.features.AbstractAlternativeFeature;
import org.fortiss.variability.model.features.AbstractCompositionalFeature;
import org.fortiss.variability.model.features.AbstractCrossFeatureConstraint;
import org.fortiss.variability.model.features.AbstractFeature;
import org.fortiss.variability.model.features.AbstractFeatureModel;
import org.fortiss.variability.model.presence.AndPC;
import org.fortiss.variability.model.presence.DefaultPC;
import org.fortiss.variability.model.presence.ILiteralReferencable;
import org.fortiss.variability.model.presence.LiteralPC;
import org.fortiss.variability.model.presence.NotPC;
import org.fortiss.variability.model.presence.OrPC;
import org.fortiss.variability.model.presence.PresenceCondition;
import org.fortiss.variability.model.presence.PresenceConditionTerm;

import com.microsoft.z3.BoolExpr;
import com.microsoft.z3.BoolSort;
import com.microsoft.z3.Context;
import com.microsoft.z3.EnumSort;
import com.microsoft.z3.Expr;
import com.microsoft.z3.FuncDecl;
import com.microsoft.z3.IntSort;
import com.microsoft.z3.SeqExpr;
import com.microsoft.z3.SeqSort;
import com.microsoft.z3.Solver;
import com.microsoft.z3.Sort;
import com.microsoft.z3.StringSymbol;
import com.microsoft.z3.Symbol;

/**
 * Base class that implements the translation of EMF based model product-lines to SMT in order to
 * perform product-line analysis via constraint lifting.
 * 
 * @author bayha
 */
public class EMFProductLineTranslation implements IProductLineTranslation {

	/** Prefix to be displayed as prefix to all logging messages from the translation. */
	private static final String LOG_PREFIX = "[EMF product-line translation]: ";

	/** {@link Context} to capture the translated model. */
	private Context ctx;

	/** Z3 sort for base type integer. */
	private IntSort intSort;
	/** Z3 sort for base type string. */
	private SeqSort<?> stringSort;
	/** Z3 sort for base type boolean. */
	private BoolSort boolSort;

	/** List of all translated assertions for the model. */
	private List<BoolExpr> modelAssertions;
	/** List of all assertions for the feature model. */
	private List<BoolExpr> featureModelAssertions;
	/** List of all assertions which encode the presence conditions. */
	private List<BoolExpr> presenceConditionAssertions;
	/** List of all assertions which encode the constraints. */
	private List<BoolExpr> constraintAssertions;

	/** Mapping of EDataTypes (e.g. EInt) to their translated Z3 type. */
	private Map<EDataType, Sort> eDataType2Sort;
	/** Mapping for classes to their respective Z3 type. */
	private Map<EClass, EnumSort<?>> eClass2z3Sort;

	/** Mapping of enumeration literals to the translated Z3 expression. */
	private Map<Enum<?>, Expr<?>> eLiteral2Expr;
	/** Mapping of classes to their respective z3 expression for 'null' */
	private Map<EClass, Expr<?>> eClass2NullElement;
	/** Mapping of enumerations to their respective z3 expression for 'null' */
	private Map<EEnum, Expr<?>> eEnum2NullElement;

	/** Mapping of attributes to their translated Z3 function. */
	private DualKeyMap<EAttribute, EClass, FuncDecl<?>> eAttribute2FunDecl =
			new DualKeyMap<EAttribute, EClass, FuncDecl<?>>();
	/** Mapping of references to their translated Z3 function. */
	private DualKeyMap<EReference, EClass, FuncDecl<?>> eReference2FunDecl =
			new DualKeyMap<EReference, EClass, FuncDecl<?>>();

	/**
	 * Mapping of objects with types to their translated Z3 expressions. The second key EClass is
	 * necessary as an object is translated for all classes specified in 'translatedClasses').
	 */
	private DualKeyMap<EObject, EClass, Expr<?>> eObject2z3Expr =
			new DualKeyMap<EObject, EClass, Expr<?>>();

	/** Mapping of classes to the selection function for this type. */
	private Map<EClass, FuncDecl<?>> eClass2SelectionFunction;

	/** The set to specify all {@link EClass} to be translated by this translation. */
	private Set<EClass> translatedClasses;
	/** The set to specify all {@link EReference} to be translated by this translation. */
	private Set<EReference> translatedReferences;
	/** The set to specify all {@link EAttribute} to be translated by this translation. */
	private Set<EAttribute> translatedAttributes;

	/** The mapping of translated {@link EClass} to the objects that have been collected. */
	private BucketSetMap<EClass, EObject> class2TranslatedEObjects;

	/** The {@link AbstractFeatureModel} collected by this translation */
	private AbstractFeatureModel featureModel;
	/** Mapping of features to corresponding Z3 {@link BoolExpr}. */
	private Map<AbstractFeature, BoolExpr> feature2BoolExpr;

	/** The mapping of translated {@link EObject}s to global {@link PresenceCondition}s. */
	private Map<EObject, PresenceConditionTerm> object2presenceCondition;

	/**
	 * The mapping from Z3 {@link Expr} back to the respective {@link EObject} to be used for
	 * interpreting solver results.
	 */
	private Map<Expr<?>, EObject> expr2EObject = new HashMap<Expr<?>, EObject>();

	/**
	 * The mapping of translated {@link EReference}s to the type that will be used for translation.
	 * This type might be different from the references original type in case only a subtype is
	 * specified for translation.
	 */
	private Map<EReference, EClass> reference2TranslatedEClass;

	/** Flag that specifies whether the optimization to only translate optional parts is enabled. */
	private boolean isIslandOptimizationEnabled = false;

	/**
	 * The number of non-variable elements that are connected to a variable element to be translated
	 * in case the "island optimization" is enabled.
	 */
	private int islandOptimizationStepNum = 0;

	/**
	 * Constructor.
	 */
	public EMFProductLineTranslation() {
		// Fields which should not change between individual translations are initialized only once
		translatedClasses = new HashSet<EClass>();
		translatedReferences = new HashSet<EReference>();
		translatedAttributes = new HashSet<EAttribute>();
	}

	/**
	 * Initialization of analysis and base types.
	 */
	private void initialize() {
		ctx = new Context();

		class2TranslatedEObjects = new BucketSetMap<EClass, EObject>();
		object2presenceCondition = new HashMap<EObject, PresenceConditionTerm>();

		reference2TranslatedEClass = new HashMap<EReference, EClass>();

		modelAssertions = new ArrayList<BoolExpr>();
		featureModelAssertions = new ArrayList<BoolExpr>();
		presenceConditionAssertions = new ArrayList<BoolExpr>();
		constraintAssertions = new ArrayList<BoolExpr>();

		eDataType2Sort = new HashMap<EDataType, Sort>();
		eClass2z3Sort = new HashMap<EClass, EnumSort<?>>();

		eLiteral2Expr = new HashMap<Enum<?>, Expr<?>>();
		eClass2NullElement = new HashMap<EClass, Expr<?>>();
		eEnum2NullElement = new HashMap<EEnum, Expr<?>>();

		eClass2SelectionFunction = new HashMap<EClass, FuncDecl<?>>();
		feature2BoolExpr = new HashMap<AbstractFeature, BoolExpr>();

		// Initialize model independent base types
		intSort = ctx.mkIntSort();
		stringSort = ctx.mkStringSort();
		boolSort = ctx.mkBoolSort();

		eDataType2Sort.put(EcorePackage.Literals.EINT, intSort);
		eDataType2Sort.put(EcorePackage.Literals.ESTRING, stringSort);
		eDataType2Sort.put(EcorePackage.Literals.EBOOLEAN, boolSort);
	}

	/** {@inheritDoc} */
	@Override
	public final void translateModel(EObject model) {
		// Initialize and reset local fields
		initialize();

		// Reference types need to be collected first, since they are required by optimizations in
		// collectTranslatedObjects().
		collectReferenceTypes();
		collectTranslatedObjects(model, null);

		translateFeatureModel();

		// Metamodel translation
		translateClassesMetamodel();
		translateAttributesAndReferencesMetamodel();

		// Model translation
		translateObjects();

		// Translate variability
		translatePresenceConditions();
	}

	/**
	 * 
	 */
	private void collectReferenceTypes() {
		for(EReference ref : translatedReferences) {
			EClass refType = ref.getEReferenceType();

			if(translatedClasses.contains(refType)) {
				reference2TranslatedEClass.put(ref, refType);
			} else {
				// Find a super type of the reference class
				for(EClass ecls : translatedClasses) {
					if(ecls.isSuperTypeOf(refType)) {
						reference2TranslatedEClass.put(ref, ecls);
					} else if(refType.isSuperTypeOf(ecls)) {
						// Attempt to use nearest found class
						reference2TranslatedEClass.put(ref, ecls);

						logInfo(LOG_PREFIX + "The reference \"" + ref.getName() + "\" of type " +
								refType.getName() +
								" is supposed to be transletd for a prodct-line analysis. Since this type is not specified to be translated, the translated sub-class " +
								ecls.getName() + " will be used.");
					}
				}

				// Check if any super type was found
				if(reference2TranslatedEClass.get(ref) == null) {
					String errMsg = LOG_PREFIX + "ERefernce \"" + ref.getName() + "\" is of type " +
							ref.getEReferenceType().getName() + " which is not translated.";

					logError(errMsg);

					throw new RuntimeException(errMsg);
				}
			}
		}
	}

	/**
	 * @param model
	 * @param parentCondition
	 */
	private void collectTranslatedObjects(EObject model, PresenceConditionTerm parentCondition) {
		if(model instanceof AbstractFeatureModel) {
			featureModel = (AbstractFeatureModel)model;

			return;
		}

		EClass modelClass = model.eClass();
		PresenceConditionTerm pc = getLocalPresenceCondition(model);

		if(pc == null) {
			pc = copy(parentCondition);
		} else {
			if(parentCondition != null)
				pc = createAndPC(copy(pc), copy(parentCondition), "");
		}

		for(EClass cls : translatedClasses) {
			if(cls.isSuperTypeOf(modelClass)) {
				// For the island optimization, only optional objects are translated.
				if(isIslandOptimizationEnabled && pc != null) {
					collectEObject(model, cls, pc);
				}
			}
		}

		for(EObject child : model.eContents()) {
			collectTranslatedObjects(child, pc);
		}
	}

	/**
	 * @param eo
	 * @param cls
	 * @param pc
	 */
	private void collectEObject(EObject eo, EClass cls, PresenceConditionTerm pc) {
		class2TranslatedEObjects.add(cls, eo);

		if(isIslandOptimizationEnabled) {
			collectTranslatedReferences(eo, islandOptimizationStepNum);
		}

		object2presenceCondition.put(eo, pc);
	}

	/**
	 * @param model
	 * @param remainingSteps
	 */
	private void collectTranslatedReferences(EObject model, int remainingSteps) {
		if(remainingSteps <= 0) {
			return;
		}

		// Collect outgoing references
		for(EReference ref : model.eClass().getEAllReferences()) {
			if(translatedReferences.contains(ref)) {
				EObject refObj = (EObject)model.eGet(ref);
				EClass refClass = reference2TranslatedEClass.get(ref);
				class2TranslatedEObjects.add(refClass, refObj);

				collectTranslatedReferences(refObj, remainingSteps - 1);
			}
		}

		// Collect incoming references
		for(Setting s : getAllReferences(model)) {
			EStructuralFeature ref = s.getEStructuralFeature();
			if(ref instanceof EReference) {
				if(translatedReferences.contains(ref)) {
					EObject referencingEObject = s.getEObject();
					EClass referencingEClass = referencingEObject.eClass();

					if(translatedClasses.contains(referencingEClass)) {
						class2TranslatedEObjects.add(referencingEClass, referencingEObject);

						collectTranslatedReferences(referencingEObject, remainingSteps - 1);

						continue;
					}

					for(EClass cls : translatedClasses) {
						if(cls.isSuperTypeOf(referencingEClass)) {
							class2TranslatedEObjects.add(cls, referencingEObject);

							collectTranslatedReferences(referencingEObject, remainingSteps - 1);
						}
					}
				}
			}
		}
	}

	/** Translates all presence conditions to Z3. */
	private void translatePresenceConditions() {
		for(EClass ec : translatedClasses) {
			EnumSort<?> ecSort = eClass2z3Sort.get(ec);
			FuncDecl<?> selFun = ctx.mkFuncDecl("SELECTED_" + ec.getName(), ecSort, boolSort);
			eClass2SelectionFunction.put(ec, selFun);

			for(EObject eo : class2TranslatedEObjects.get(ec)) {
				BoolExpr eoSelection;

				final PresenceConditionTerm globalPC = object2presenceCondition.get(eo);
				if(globalPC != null) {
					eoSelection = translatePresenceCondition(globalPC);
				} else {
					eoSelection = ctx.mkTrue();
				}

				Expr<?> eoExpr = eObject2z3Expr.get(eo, ec);
				presenceConditionAssertions.add(ctx.mkEq(ctx.mkApp(selFun, eoExpr), eoSelection));
			}
		}
	}

	/** Translates a given {@link PresenceCondition} to Z3. */
	private BoolExpr translatePresenceCondition(PresenceCondition pc) {
		if(pc == null) {
			return null;
		}

		if(pc instanceof LiteralPC) {
			ILiteralReferencable literal = ((LiteralPC)pc).getLiteralReference();
			if(literal == null || literal.getName().length() == 0) {
				// Illegal literals are translated as true.
				return ctx.mkTrue();
			}
			return feature2BoolExpr.get(literal);
		}

		if(pc instanceof DefaultPC) {
			return translatePresenceCondition(pc.resolveToFeatureLiterals());
		}

		if(pc instanceof OrPC) {
			PresenceConditionTerm op1 = ((OrPC)pc).getOperand1();
			PresenceConditionTerm op2 = ((OrPC)pc).getOperand2();
			return ctx.mkOr(translatePresenceCondition(op1), translatePresenceCondition(op2));
		}

		if(pc instanceof AndPC) {
			PresenceConditionTerm op1 = ((AndPC)pc).getOperand1();
			PresenceConditionTerm op2 = ((AndPC)pc).getOperand2();
			return ctx.mkAnd(translatePresenceCondition(op1), translatePresenceCondition(op2));
		}

		if(pc instanceof NotPC) {
			return ctx.mkNot(translatePresenceCondition(((NotPC)pc).getOperand()));
		}

		logError(LOG_PREFIX + "Unknown PresenceCondition: " + pc);
		return null;
	}

	/** Translates the feature model to Z3. */
	private void translateFeatureModel() {
		if(featureModel == null) {
			return;
		}

		BoolExpr fmExpr = ctx.mkBoolConst("FEATURE_" + featureModel.getName());
		feature2BoolExpr.put(featureModel, fmExpr);

		featureModelAssertions.add(fmExpr);

		// Translate all features (including alternatives)
		for(AbstractFeature f : getChildrenWithType(featureModel, AbstractFeature.class)) {
			String name = f.getName();
			String constName = "FEATURE_" + name;
			final BoolExpr fExpr = ctx.mkBoolConst(constName.replaceAll("@", "AT"));
			feature2BoolExpr.put(f, fExpr);

			// If f is child of another feature, the selection criteria have to be encoded.
			final EObject fContainer = f.eContainer();
			if(fContainer instanceof AbstractFeature) {
				AbstractFeature parentF = (AbstractFeature)fContainer;
				final BoolExpr parentExpr = feature2BoolExpr.get(parentF);

				if(f.isOptional()) {
					featureModelAssertions.add(ctx.mkImplies(fExpr, parentExpr));
				} else {
					if(fContainer instanceof AbstractCompositionalFeature) {
						featureModelAssertions.add(ctx.mkImplies(parentExpr, fExpr));
					}
				}
			}
		}

		// For alternatives also translate the alternative selection criteria
		for(AbstractAlternativeFeature af : getChildrenWithType(featureModel,
				AbstractAlternativeFeature.class)) {
			BoolExpr afExpr = feature2BoolExpr.get(af);
			EList<AbstractFeature> as = af.getAlternatives();

			featureModelAssertions.add(
					ctx.mkImplies(afExpr, ctx.mkOr(as.stream().map(alt -> feature2BoolExpr.get(alt))
							.collect(toList()).toArray(new BoolExpr[0]))));

			for(AbstractFeature a : as) {
				BoolExpr aExpr = feature2BoolExpr.get(a);
				featureModelAssertions.add(ctx.mkImplies(aExpr, afExpr));
				featureModelAssertions.add(ctx.mkImplies(aExpr,
						ctx.mkAnd(as.stream().filter(alt -> alt != a)
								.map(alt -> ctx.mkNot(feature2BoolExpr.get(alt))).collect(toList())
								.toArray(new BoolExpr[0]))));
			}
		}

		// Translate all constraints.
		for(AbstractCrossFeatureConstraint c : getChildrenWithType(featureModel,
				AbstractCrossFeatureConstraint.class)) {
			AbstractFeature target = c.getTarget();
			AbstractFeature source = (AbstractFeature)c.eContainer();

			BoolExpr srcExpr = feature2BoolExpr.get(source);
			BoolExpr trgExpr = feature2BoolExpr.get(target);

			switch(c.getType()) {
				case REQUIRES:
					featureModelAssertions.add(ctx.mkImplies(srcExpr, trgExpr));
					break;
				case EXCLUDES:
					featureModelAssertions.add(ctx.mkImplies(srcExpr, ctx.mkNot(trgExpr)));
					break;
				default: // Suggestions are not translated.
			}
		}
	}

	/** Translates all objects in the model. */
	private void translateObjects() {
		// Translate objects by class.
		for(EClass ec : translatedClasses) {
			Set<EObject> objs = class2TranslatedEObjects.get(ec);

			EList<EAttribute> attributes = ec.getEAllAttributes();
			for(EAttribute ea : attributes) {
				if(translatedAttributes.contains(ea)) {
					FuncDecl<?> attFunc = eAttribute2FunDecl.get(ea, ec);

					for(EObject eo : objs) {
						Expr<?> attVal = translateAttributeValue(eo, ea);
						BoolExpr attValAssertion =
								ctx.mkEq(ctx.mkApp(attFunc, eObject2z3Expr.get(eo, ec)), attVal);
						modelAssertions.add(attValAssertion);
					}
				}
			}

			EList<EReference> references = ec.getEAllReferences();
			for(EReference er : references) {
				if(translatedReferences.contains(er)) {
					FuncDecl<?> refFunc = eReference2FunDecl.get(er, ec);
					for(EObject eo : objs) {
						Expr<?> eoExpr = eObject2z3Expr.get(eo, ec);
						Expr<?> refVal = translateReferenceValue(eo, er);
						BoolExpr rafValAssertion = ctx.mkEq(ctx.mkApp(refFunc, eoExpr), refVal);
						modelAssertions.add(rafValAssertion);
					}

					// Null must always only point to null....
					EClass refType = reference2TranslatedEClass.get(er);
					Expr<?> nullElement = eClass2NullElement.get(ec);
					if(er.getUpperBound() > 0) {
						modelAssertions.add(ctx.mkEq(ctx.mkApp(refFunc, nullElement),
								eClass2NullElement.get(refType)));
					} else {
						modelAssertions
								.add(ctx.mkEq(ctx.mkApp(refFunc, nullElement), ctx.mkEmptySeq(
										eClass2z3Sort.get(reference2TranslatedEClass.get(er)))));
					}
				}
			}
		}
	}

	/** Translates the given instance of a relation between individual objects in the model. */
	@SuppressWarnings({"unchecked", "rawtypes"})
	private Expr<?> translateReferenceValue(EObject eo, EReference er) {
		Object value = eo.eGet(er);
		EClass refType = reference2TranslatedEClass.get(er);
		Expr<?> refNullValue = eClass2NullElement.get(refType);

		FuncDecl<?> refSelFun = eClass2SelectionFunction.get(refType);

		if(er.getUpperBound() < 0) {
			// Case for lists
			List<?> values = (List<?>)value;

			if(values.size() == 0) {
				return ctx.mkEmptySeq(getSortForReference(er));
			}

			SeqExpr[] seqElems = new SeqExpr[values.size()];
			for(int i = 0; i < values.size(); i++) {
				Object o = values.get(i);

				Expr<?> elemExpr = eObject2z3Expr.get((EObject)o, refType);

				SeqExpr<?> selectedExpr = ctx.mkUnit(elemExpr);
				if(refSelFun != null) {
					SeqExpr notSelectedExpr = ctx.mkEmptySeq(getSortForReference(er));
					seqElems[i] = (SeqExpr<?>)ctx.mkITE((BoolExpr)ctx.mkApp(refSelFun, elemExpr),
							selectedExpr, notSelectedExpr);
				} else {
					seqElems[i] = selectedExpr;
				}
			}

			if(seqElems.length == 0) {
				return ctx.mkEmptySeq(getSortForReference(er));
			}
			if(seqElems.length == 1) {
				return seqElems[0];
			}
			return ctx.mkConcat(seqElems);
		}

		if(value == null) {
			return refNullValue;
		}

		Expr singleValExpr = eObject2z3Expr.get((EObject)value, refType);

		if(refSelFun != null) {
			// Case with variable target
			BoolExpr eoSel = (BoolExpr)ctx.mkApp(refSelFun, singleValExpr);

			return ctx.mkITE(eoSel, singleValExpr, refNullValue);
		}

		// Default case with non variable target.
		return singleValExpr;
	}

	/** Translates the given EAttribute value for the given object to Z3. */
	@SuppressWarnings({"rawtypes", "unchecked"})
	private Expr<?> translateAttributeValue(EObject eo, EAttribute ea) {
		Object value = eo.eGet(ea);
		EDataType attType = ea.getEAttributeType();

		if(ea.getUpperBound() < 0 && value instanceof List<?>) {
			// Case for lists.
			List<?> values = (List<?>)value;

			SeqExpr[] seqElem = new SeqExpr[values.size()];
			for(int i = 0; i < values.size(); i++) {
				Object o = values.get(i);
				Expr<?> elemExpr = translatePrimitiveObject(o, attType);

				seqElem[i] = ctx.mkUnit(elemExpr);
			}

			return ctx.mkConcat(seqElem);
		}

		return translatePrimitiveObject(value, attType);
	}

	/** Translates the given object for a primitive type (Enums, Integer, String, Boolean) to Z3. */
	private Expr<?> translatePrimitiveObject(Object value, EDataType attType) {
		// Single value
		if(attType.equals(EcorePackage.Literals.EINT)) {
			if(value == null) {
				logError(LOG_PREFIX + "null int found");
				return ctx.mkInt(0);
			}
			return ctx.mkInt((int)value);
		}
		if(attType.equals(EcorePackage.Literals.ESTRING)) {
			if(value == null) {
				return ctx.mkString("");
			}
			return ctx.mkString((String)value);
		}
		if(attType.equals(EcorePackage.Literals.EBOOLEAN)) {
			if(value == null) {
				logError(LOG_PREFIX + "null bool found");
				return ctx.mkBool(false);
			}
			return ctx.mkBool((boolean)value);
		}
		if(attType instanceof EEnum) {
			if(value == null) {
				return eEnum2NullElement.get(attType);
			}
			return eLiteral2Expr.get(value);
		}

		return null;
	}

	/** Creates functions for all attributes and references of the meta model in Z3. */
	private void translateAttributesAndReferencesMetamodel() {
		for(EClass ec : translatedClasses) {
			for(EAttribute ea : ec.getEAllAttributes()) {
				if(translatedAttributes.contains(ea)) {
					String funId = ec.getName() + "_" + ea.getName();

					EDataType attDataType = ea.getEAttributeType();
					Sort rangeSort = eDataType2Sort.get(attDataType);
					if(attDataType instanceof EEnum) {
						if(rangeSort == null) {
							rangeSort = translateEnum((EEnum)attDataType);
							eDataType2Sort.put(attDataType, rangeSort);
						}
					}

					if(rangeSort == null) {
						// Attribute is not translated, as all attribute types except enums are
						// known.
						continue;
					}

					FuncDecl<?> funDecl = ctx.mkFuncDecl(funId, eClass2z3Sort.get(ec), rangeSort);

					eAttribute2FunDecl.put(ea, ec, funDecl);
				}
			}

			for(EReference er : ec.getEAllReferences()) {
				if(translatedReferences.contains(er)) {
					String funId = ec.getName() + "_" + er.getName();

					EClass refType = reference2TranslatedEClass.get(er);

					Sort refSort = eClass2z3Sort.get(refType);

					if(refSort == null) {
						// Find translated subclass of reference target and attempt to use this.
						for(EClass ecls : translatedClasses) {
							if(refType.isSuperTypeOf(ecls)) {
								refSort = eClass2z3Sort.get(ecls);

								logInfo(LOG_PREFIX + "The reference \"" + er.getName() +
										"\" of type " + refType.getName() +
										" is supposed to be transletd for a prodct-line analysis. Since this type is not specified to be translated, the translated sub-class " +
										ecls.getName() + " will be used.");
							}
						}
					}

					if(refSort == null) {
						String msg = "Reference " + er.getName() +
								" shall be translated, but is of type " + refType.getName() +
								" which is not translated. Add the type to the translated classes to translate this reference.";

						logError(LOG_PREFIX + msg, null);
						throw new RuntimeException(msg);
					}

					if(er.getUpperBound() < 0) {
						// This is the case of unlimited multiplicity, i.e. lists
						refSort = ctx.mkSeqSort(refSort);
					}

					FuncDecl<?> funDecl = ctx.mkFuncDecl(funId, eClass2z3Sort.get(ec), refSort);

					eReference2FunDecl.put(er, ec, funDecl);
				}
			}
		}
	}

	/** Translates the give enumeration */
	private Sort translateEnum(EEnum en) {
		String name = en.getName();
		EList<EEnumLiteral> enumLiterals = en.getELiterals();

		// Null needs to be encoded as a dedicated literal
		int numLiterals = enumLiterals.size();
		String[] literals = new String[numLiterals + 1];
		for(int i = 0; i < numLiterals; i++) {
			literals[i] = name + "_" + enumLiterals.get(i).getName();
		}
		literals[numLiterals] = name + "_NONE_LITERAL";

		EnumSort<?> enumSort = ctx.mkEnumSort(name, literals);

		// Add regular literals and null literal to the respective mappings.
		for(int i = 0; i < numLiterals; i++) {
			eLiteral2Expr.put((Enum<?>)enumLiterals.get(i).getInstance(), enumSort.getConst(i));
		}
		eEnum2NullElement.put(en, enumSort.getConst(numLiterals));

		return enumSort;
	}

	/** Creates the types for all classes and objects in Z3. */
	private void translateClassesMetamodel() {
		for(EClass cls : translatedClasses) {
			List<EObject> objs = new ArrayList<EObject>(class2TranslatedEObjects.get(cls));

			// The last index is added for the null element.
			Symbol[] objSymbols = new Symbol[objs.size() + 1];
			String clsName = cls.getName();
			for(int i = 0; i < objs.size(); i++) {
				String id = clsName + i;

				StringSymbol objSymbol = ctx.mkSymbol(id);
				objSymbols[i] = objSymbol;
			}
			objSymbols[objs.size()] = ctx.mkSymbol(clsName + "NONE");

			// Create EnumSort for class
			EnumSort<?> enumSort = ctx.mkEnumSort(ctx.mkSymbol(clsName), objSymbols);
			eClass2z3Sort.put(cls, enumSort);

			// Map objects to z3 enum literals
			Expr<?>[] consts = enumSort.getConsts();
			for(int i = 0; i < objs.size(); i++) {
				eObject2z3Expr.put(objs.get(i), cls, consts[i]);
				expr2EObject.put(consts[i], objs.get(i));
			}
			Expr<?> noneExpr = consts[consts.length - 1];
			eClass2NullElement.put(cls, noneExpr);
		}
	}

	/** {@inheritDoc} */
	@Override
	public Solver createSolver() {
		Solver solver = ctx.mkSolver();

		solver.add(featureModelAssertions.toArray(new BoolExpr[featureModelAssertions.size()]));
		solver.add(presenceConditionAssertions
				.toArray(new BoolExpr[presenceConditionAssertions.size()]));
		solver.add(modelAssertions.toArray(new BoolExpr[modelAssertions.size()]));

		solver.add(constraintAssertions.toArray(new BoolExpr[constraintAssertions.size()]));

		return solver;
	}

	/** {@inheritDoc} */
	@Override
	public Context getContext() {
		return ctx;
	}

	/** {@inheritDoc} */
	@Override
	public EObject getEObjectForExpression(Expr<?> expr) {
		return expr2EObject.get(expr);
	}

	/** {@inheritDoc} */
	@Override
	public Expr<?> getNullElement(EClass cls) {
		return eClass2NullElement.get(cls);
	}

	/** {@inheritDoc} */
	@Override
	public EnumSort<?> getDatatype(EClass ec) {
		return eClass2z3Sort.get(ec);
	}

	/** {@inheritDoc} */
	@Override
	public FuncDecl<?> getAttributeReferenceFunctionDeclaration(EClass cVarType, String name) {
		for(EReference ref : cVarType.getEAllReferences()) {
			if(ref.getName().equals(name)) {
				return eReference2FunDecl.get(ref, cVarType);
			}
		}

		for(EAttribute att : cVarType.getEAllAttributes()) {
			if(att.getName().equals(name)) {
				return eAttribute2FunDecl.get(att, cVarType);
			}
		}

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Map<BoolExpr, IProductLineConstraint>
			translateConstraints(Collection<IProductLineConstraint> constraints) {
		Map<BoolExpr, IProductLineConstraint> constraintTracker2Constraint =
				new HashMap<BoolExpr, IProductLineConstraint>();

		for(IProductLineConstraint constraint : constraints) {
			Expr<?>[] quantifierValiables = constraint.getQuantifierVariables();
			EClass[] variableClasses = constraint.getVariablesClasses();
			BoolExpr[] containmentExprs = new BoolExpr[0];
			BoolExpr body = constraint.getBody();

			boolean isForAll = constraint.isForAll();

			BoolExpr quantifier = createLiftedConstraint(quantifierValiables, variableClasses,
					containmentExprs, body, isForAll);

			BoolExpr constraintTracker =
					ctx.mkBoolConst("CONSTRAINT_" + constraint.getConstraintName());
			constraintTracker2Constraint.put(constraintTracker, constraint);

			constraintAssertions.add(ctx.mkEq(constraintTracker, quantifier));
		}

		Set<BoolExpr> allConstraints = constraintTracker2Constraint.keySet();
		constraintAssertions.add(ctx.mkOr(allConstraints.toArray(new BoolExpr[0])));

		return constraintTracker2Constraint;
	}

	/** Automatically lifts the given constraint */
	protected BoolExpr createLiftedConstraint(Expr<?>[] quantifierValiables,
			EClass[] variableClasses, BoolExpr[] containmentExprs, BoolExpr body,
			boolean isForAll) {

		BoolExpr[] selExprs = new BoolExpr[quantifierValiables.length + containmentExprs.length];
		for(int i = 0; i < quantifierValiables.length; i++) {
			EClass cls = variableClasses[i];
			Expr<?> var = quantifierValiables[i];

			FuncDecl<?> selFun = eClass2SelectionFunction.get(cls);
			BoolExpr selected = (BoolExpr)ctx.mkApp(selFun, var);
			BoolExpr notNull = ctx.mkNot(ctx.mkEq(var, getNullElement(cls)));

			selExprs[i] = ctx.mkAnd(selected, notNull);
		}

		// Add containment expressions
		int j = 0;
		for(int i = quantifierValiables.length; i < quantifierValiables.length +
				containmentExprs.length; i++) {
			selExprs[i] = containmentExprs[j++];
		}

		if(isForAll) {
			BoolExpr liftedBody = ctx.mkImplies(ctx.mkAnd(selExprs), body);
			return ctx.mkForall(quantifierValiables, liftedBody, 0, null, null, null, null);
		}

		// else: Exits
		BoolExpr liftedBody = ctx.mkAnd(ctx.mkAnd(selExprs), body);
		return ctx.mkExists(quantifierValiables, liftedBody, 0, null, null, null, null);
	}

	/** Utility method to get the z3 sort for a given reference. */
	private EnumSort<?> getSortForReference(EReference er) {
		return eClass2z3Sort.get(reference2TranslatedEClass.get(er));
	}

	/**
	 * Retrieves the {@link PresenceCondition} that was assigned to 'modelElement' in the model.
	 * This function might be overwritten in case an alternative location for presence conditions
	 * was defined.
	 */
	protected PresenceConditionTerm getLocalPresenceCondition(EObject model) {
		if(model instanceof IOptionalVariationPoint<?>) {
			PresenceCondition presenceCondition =
					((IOptionalVariationPoint<?>)model).getPresenceCondition();

			if(presenceCondition == null) {
				return null;
			}

			return presenceCondition.resolveToFeatureLiterals();
		}
		if(model instanceof IAlternative) {
			return ((IAlternative)model).getPresenceCondition().resolveToFeatureLiterals();
		}

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void setTranslatedEClasses(Collection<EClass> clss) {
		translatedClasses.clear();

		translatedClasses.addAll(clss);
	}

	/** {@inheritDoc} */
	@Override
	public void setTranslatedEReferences(Collection<EReference> refs) {
		translatedReferences.clear();

		translatedReferences.addAll(refs);
	}

	/** {@inheritDoc} */
	@Override
	public void setTranslatedEAttributes(Collection<EAttribute> atts) {
		translatedAttributes.clear();

		translatedAttributes.addAll(atts);
	}

	/**
	 * Enables or disables the "island optimization" for which only optional parts of the model are
	 * translated.
	 * This optimization is disbaled by default.
	 * 
	 * Via the method {@code setIslandOptimizationStepNum()} it is possible to specify a range of
	 * transitive references to be guaranteed to be translated for all optional elements.
	 * Note, that this step size might be crucial for a constraint to checked correctly for
	 * interactions between optional and non-optional elements.
	 * 
	 * @param setEnabled
	 *            'true' will enable the island optimization; 'false' will disbale.
	 */
	public void setIslandOptimizationEnabled(boolean setEnabled) {
		this.isIslandOptimizationEnabled = setEnabled;
	}

	/**
	 * Specifies a number of transitive references starting from every optional element that will be
	 * guaranteed to be translated - even if they are non-optional.
	 * 
	 * Note, that this step size might be crucial for a constraint to checked correctly for
	 * interactions between optional and non-optional elements.
	 * 
	 * @param stepNum
	 *            The number of trabnsitive references to be guaranteed to be translated for all
	 *            optional elements.
	 */
	public void setIslandOptimizationStepNum(int stepNum) {
		this.islandOptimizationStepNum = stepNum;
	}
}
