package org.fortiss.variability.analysis;

import static com.microsoft.z3.Status.UNSATISFIABLE;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.fortiss.tooling.kernel.model.INamedElement;

import com.microsoft.z3.BoolExpr;
import com.microsoft.z3.Expr;
import com.microsoft.z3.Model;
import com.microsoft.z3.Solver;
import com.microsoft.z3.Status;

/**
 * Base class that implements the translation of model product-lines to SMT in order to perform
 * product-line analysis via constraint lifting.
 * 
 * @author bayha
 */
public abstract class GenericProductLineAnalysis {

	protected final IProductLineTranslation translation;

	protected List<IProductLineConstraint> plConstraints = new ArrayList<IProductLineConstraint>();

	protected final EObject model;

	/**
	 * Constructor.
	 */
	protected GenericProductLineAnalysis(EObject model, IProductLineTranslation translation) {
		this.model = model;

		this.translation = translation;
	}

	/**
	 * Performs a product-line analysis for the given model.
	 * 
	 * @return 'true' if all constraints are fulfilled for all variants of the product-line
	 *         analysis. 'false' otherwise.
	 */
	public List<ProductLineConstraintViolation> doCheck() {
		final long start = System.currentTimeMillis();

		addConstriants();

		setTranslatedClassesAndReferences();

		translation.translateModel(model);

		Map<BoolExpr, IProductLineConstraint> constraintTracker2Constraint =
				translation.translateConstraints(plConstraints);

		// addRawConstraints();

		// Global.setParameter("parallel.enable", "true");

		Solver solver = translation.createSolver();

		String modelName = ((INamedElement)model).getName();
		String path = "C:\\Users\\bayha\\tmp\\z3" + modelName + (start % 159739431) + "in";
		try {
			PrintWriter printWriter = new PrintWriter(path);
			printWriter.println(solver);
			printWriter.println("(check-sat)");
			// printWriter.println("(get-model)");
			printWriter.close();
		} catch(FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		long startSmt = System.currentTimeMillis();

		Status result = solver.check();

		long end = System.currentTimeMillis();

		System.out.println("Runtime " + modelName + ": " + (end - start) + " total, " +
				(startSmt - start) + " translation, " + (end - startSmt) + " SMT");

		List<ProductLineConstraintViolation> ret = new ArrayList<ProductLineConstraintViolation>();

		if(result.equals(UNSATISFIABLE)) {
			// No constraint was violated

			// BoolExpr[] unsatCore = solver.getUnsatCore();
			//
			// PrintWriter printWriter;
			//
			// try {
			// printWriter = new PrintWriter(
			// "C:\\Users\\bayha\\tmp\\z3" + modelName + (start % 159739431) + "unsat");
			//
			// for(BoolExpr b : unsatCore) {
			// BoolExpr coreExpr = tracToExpr.get(b);
			//
			// printWriter.println(coreExpr);
			// }
			//
			// printWriter.close();
			// } catch(FileNotFoundException e) {
			// // TODO Auto-generated catch block
			// e.printStackTrace();
			// }

			// System.out.println(deepToString(unsatCore));
		} else {

			Model z3Model = solver.getModel();

			for(BoolExpr constTrack : constraintTracker2Constraint.keySet()) {
				Expr constInterp = z3Model.getConstInterp(constTrack);

				if(constInterp instanceof BoolExpr) {
					if(constInterp.equals(translation.getContext().mkTrue())) {
						IProductLineConstraint failedConstraint =
								constraintTracker2Constraint.get(constTrack);

						List<EObject> violatingObjects = new ArrayList<EObject>();
						for(Expr varTrack : failedConstraint.getVariableTracker()) {
							Expr interpr = z3Model.getConstInterp(varTrack);
							EObject eObject = translation.getEObjectForExpression(interpr);

							violatingObjects.add(eObject);
						}

						ret.add(new ProductLineConstraintViolation(failedConstraint,
								violatingObjects, null));
					}
				}

			}

			try {
				PrintWriter printWriter = new PrintWriter(
						"C:\\Users\\bayha\\tmp\\z3" + modelName + (start % 159739431) + "out");
				printWriter.println(z3Model);
				printWriter.close();
			} catch(FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		translation.getContext().close();

		return ret;
	}

	/**
	 * 
	 */
	private void setTranslatedClassesAndReferences() {
		Set<EClass> classes = new HashSet<EClass>();
		Set<EReference> references = new HashSet<EReference>();
		Set<EAttribute> attributes = new HashSet<EAttribute>();

		for(IProductLineConstraint constraint : plConstraints) {
			classes.addAll(constraint.getTranslatedClasses());
			references.addAll(constraint.getTranslatedReferences());
			attributes.addAll(constraint.getTranslatedAttributes());
		}

		translation.setTranslatedEClasses(classes);
		translation.setTranslatedEReferences(references);
		translation.setTranslatedEAttributes(attributes);

	}

	protected abstract void addConstriants();
}
